@@ -22,12 +22,12 @@ import logging
 import os
 
 from keystoneclient import access
-from keystoneclient.auth.identity.base import BaseIdentityPlugin
+from keystoneclient import adapter
 import requests
 
 from neutronclient.common import exceptions
 from neutronclient.common import utils
-from neutronclient.openstack.common.gettextutils import _
+from neutronclient.i18n import _
 
 _logger = logging.getLogger(__name__)
 
@@ -42,25 +42,25 @@ else:
 logging.getLogger("requests").setLevel(_requests_log_level)
 
 
-class NeutronClientMixin(object):
+class HTTPClient(object):
+    """Handles the REST calls and responses, include authn."""
 
     USER_AGENT = 'python-neutronclient'
+    CONTENT_TYPE = 'application/json'
 
-
-class HTTPClient(NeutronClientMixin):
-    """Handles the REST calls and responses, include authn."""
+    # 8192 Is the default max URI len for eventlet.wsgi.server
+    MAX_URI_LEN = 8192
 
     def __init__(self, username=None, user_id=None,
                  tenant_name=None, tenant_id=None,
                  password=None, auth_url=None,
                  token=None, region_name=None, timeout=None,
                  endpoint_url=None, insecure=False,
-                 endpoint_type='publicURL',
+                 endpoint_type='publicURL', auth_plugin=None,
                  auth_strategy='keystone', ca_cert=None, log_credentials=False,
                  service_type='network',
                  **kwargs):
 
-        self.roles = kwargs.get('roles', '')
         self.username = username
         self.user_id = user_id
         self.tenant_name = tenant_name
@@ -74,7 +74,7 @@ class HTTPClient(NeutronClientMixin):
         self.auth_token = token
         self.auth_tenant_id = None
         self.auth_user_id = None
-        self.content_type = 'application/json'
+        self.auth_plugin = auth_plugin
         self.endpoint_url = endpoint_url
         self.auth_strategy = auth_strategy
         self.log_credentials = log_credentials
@@ -88,17 +88,8 @@ class HTTPClient(NeutronClientMixin):
         kargs.setdefault('headers', kwargs.get('headers', {}))
         kargs['headers']['User-Agent'] = self.USER_AGENT
 
-        if 'content_type' in kwargs:
-            kargs['headers']['Content-Type'] = kwargs['content_type']
-            kargs['headers']['Accept'] = kwargs['content_type']
-        else:
-            kargs['headers']['Content-Type'] = self.content_type
-            kargs['headers']['Accept'] = self.content_type
-
         if 'body' in kwargs:
             kargs['body'] = kwargs['body']
-        args = utils.safe_encode_list(args)
-        kargs = utils.safe_encode_dict(kargs)
 
         if self.log_credentials:
             log_kargs = kargs
@@ -133,28 +124,43 @@ class HTTPClient(NeutronClientMixin):
     def authenticate_and_fetch_endpoint_url(self):
         if not self.auth_token:
             self.authenticate()
-        elif not self.endpoint_url:
+        if not self.endpoint_url:
             self.endpoint_url = self._get_endpoint_url()
 
-    def request(self, url, method, **kwargs):
-        kwargs.setdefault('headers', kwargs.get('headers', {}))
-        kwargs['headers']['User-Agent'] = self.USER_AGENT
-        kwargs['headers']['Accept'] = 'application/json'
-        if 'body' in kwargs:
-            kwargs['headers']['Content-Type'] = 'application/json'
-            kwargs['data'] = kwargs['body']
-            del kwargs['body']
+    def request(self, url, method, body=None, headers=None, **kwargs):
+        """Request without authentication."""
+
+        content_type = kwargs.pop('content_type', None) or 'application/json'
+        headers = headers or {}
+        headers.setdefault('Accept', content_type)
+
+        if body:
+            headers.setdefault('Content-Type', content_type)
+
+        headers['User-Agent'] = self.USER_AGENT
+
         resp = requests.request(
             method,
             url,
+            data=body,
+            headers=headers,
             verify=self.verify_cert,
             timeout=self.timeout,
             **kwargs)
 
         return resp, resp.text
 
+    def _check_uri_length(self, action):
+        uri_len = len(self.endpoint_url) + len(action)
+        if uri_len > self.MAX_URI_LEN:
+            raise exceptions.RequestURITooLong(
+                excess=uri_len - self.MAX_URI_LEN)
+
     def do_request(self, url, method, **kwargs):
+        # Ensure client always has correct uri - do not guesstimate anything
         self.authenticate_and_fetch_endpoint_url()
+        self._check_uri_length(url)
+
         # Perform the request once. If we get a 401 back then it
         # might be because the auth token expired, so try to
         # re-authenticate and try again. If it still fails, bail.
@@ -163,19 +169,29 @@ class HTTPClient(NeutronClientMixin):
             if self.auth_token is None:
                 self.auth_token = ""
             kwargs['headers']['X-Auth-Token'] = self.auth_token
-            if self.roles is None:
-                self.roles = []
-            kwargs['headers']['X-Roles'] = ','.join(self.roles)
+            if self.auth_plugin:
+                aph = self.auth_plugin.pre_hook
+                self.endpoint_url, url, method, kwargs = aph(self.endpoint_url,
+                                                             url, method,
+                                                             **kwargs)
             resp, body = self._cs_request(self.endpoint_url + url, method,
                                           **kwargs)
+            if self.auth_plugin:
+                resp, body = self.auth_plugin.post_hook(resp, body)
             return resp, body
         except exceptions.Unauthorized:
             self.authenticate()
             kwargs.setdefault('headers', {})
-            if self.auth_strategy == 'keystone':
-                kwargs['headers']['X-Auth-Token'] = self.auth_token
+            kwargs['headers']['X-Auth-Token'] = self.auth_token
+            if self.auth_plugin:
+                aph = self.auth_plugin.pre_hook
+                self.endpoint_url, url, method, kwargs = aph(self.endpoint_url,
+                                                             url, method,
+                                                             **kwargs)
             resp, body = self._cs_request(
                 self.endpoint_url + url, method, **kwargs)
+            if self.auth_plugin:
+                resp, body = self.auth_plugin.post_hook(resp, body)
             return resp, body
 
     def _extract_service_catalog(self, body):
@@ -192,29 +208,8 @@ class HTTPClient(NeutronClientMixin):
                 service_type=self.service_type,
                 endpoint_type=self.endpoint_type)
 
-    def _authenticate_rackspace(self):
-        body = {'auth': {'RAX-KSKEY:apiKeyCredentials': {
-                         'username': self.username,
-                         'apiKey': self.auth_token}}, }
-
-        if self.auth_url is None:
-            raise exceptions.NoAuthURLProvided()
-
-        token_url = self.auth_url + "/tokens" 
-        resp, resp_body = self._cs_request(token_url, "POST",
-                                           body=json.dumps(body),
-                                           content_type="application/json",
-                                           allow_redirects=True)
-        if resp.status_code != 200:
-            raise exceptions.Unauthorized(message=resp_body)
-        if resp_body:
-            try:
-                resp_body = json.loads(resp_body)
-            except ValueError:
-                pass
-        else:
-            resp_body = None
-        self._extract_service_catalog(resp_body)
+    def _plugin_auth(self, auth_url):
+        return self.auth_plugin.authenticate(self, auth_url)
 
     def _authenticate_keystone(self):
         if self.user_id:
@@ -258,8 +253,8 @@ class HTTPClient(NeutronClientMixin):
             raise exceptions.Unauthorized(message=message)
 
     def authenticate(self):
-        if self.auth_strategy == 'rackspace':
-            self._authenticate_rackspace()
+        if self.auth_plugin:
+            self._plugin_auth(self.auth_url)
         elif self.auth_strategy == 'keystone':
             self._authenticate_keystone()
         elif self.auth_strategy == 'noauth':
@@ -284,7 +279,7 @@ class HTTPClient(NeutronClientMixin):
         body = json.loads(body)
         for endpoint in body.get('endpoints', []):
             if (endpoint['type'] == 'network' and
-                endpoint.get('region') == self.region_name):
+                    endpoint.get('region') == self.region_name):
                 if self.endpoint_type not in endpoint:
                     raise exceptions.EndpointTypeNotFound(
                         type_=self.endpoint_type)
@@ -299,75 +294,67 @@ class HTTPClient(NeutronClientMixin):
                 'endpoint_url': self.endpoint_url}
 
 
-class SessionClient(NeutronClientMixin):
+class SessionClient(adapter.Adapter):
 
-    def __init__(self,
-                 session,
-                 auth,
-                 interface=None,
-                 service_type=None,
-                 region_name=None):
-
-        self.session = session
-        self.auth = auth
-        self.interface = interface
-        self.service_type = service_type
-        self.region_name = region_name
-        self.auth_token = None
-        self.endpoint_url = None
-
-    def request(self, url, method, **kwargs):
-        kwargs.setdefault('user_agent', self.USER_AGENT)
-        kwargs.setdefault('auth', self.auth)
+    def request(self, *args, **kwargs):
         kwargs.setdefault('authenticated', False)
+        kwargs.setdefault('raise_exc', False)
+
+        content_type = kwargs.pop('content_type', None) or 'application/json'
+
+        headers = kwargs.setdefault('headers', {})
+        headers.setdefault('Accept', content_type)
 
         try:
-            kwargs['data'] = kwargs.pop('body')
+            kwargs.setdefault('data', kwargs.pop('body'))
         except KeyError:
             pass
 
-        endpoint_filter = kwargs.setdefault('endpoint_filter', {})
-        endpoint_filter.setdefault('interface', self.interface)
-        endpoint_filter.setdefault('service_type', self.service_type)
-        endpoint_filter.setdefault('region_name', self.region_name)
+        if kwargs.get('data'):
+            headers.setdefault('Content-Type', content_type)
 
-        kwargs = utils.safe_encode_dict(kwargs)
-        resp = self.session.request(url, method, **kwargs)
+        resp = super(SessionClient, self).request(*args, **kwargs)
         return resp, resp.text
 
     def do_request(self, url, method, **kwargs):
         kwargs.setdefault('authenticated', True)
         return self.request(url, method, **kwargs)
 
-    def authenticate(self):
-        # This method is provided for backward compatibility only.
-        # We only care about setting the service endpoint.
-        self.endpoint_url = self.session.get_endpoint(
-            self.auth,
-            service_type=self.service_type,
-            region_name=self.region_name,
-            interface=self.interface)
+    @property
+    def endpoint_url(self):
+        # NOTE(jamielennox): This is used purely by the CLI and should be
+        # removed when the CLI gets smarter.
+        return self.get_endpoint()
 
-    def authenticate_and_fetch_endpoint_url(self):
-        # This method is provided for backward compatibility only.
-        # We only care about setting the service endpoint.
-        self.authenticate()
+    @property
+    def auth_token(self):
+        # NOTE(jamielennox): This is used purely by the CLI and should be
+        # removed when the CLI gets smarter.
+        return self.get_token()
+
+    def authenticate(self):
+        # NOTE(jamielennox): This is used purely by the CLI and should be
+        # removed when the CLI gets smarter.
+        self.get_token()
 
     def get_auth_info(self):
-        # This method is provided for backward compatibility only.
-        if not isinstance(self.auth, BaseIdentityPlugin):
-            msg = ('Auth info not available. Auth plugin is not an identity '
-                   'auth plugin.')
-            raise exceptions.NeutronClientException(message=msg)
-        access_info = self.auth.get_access(self.session)
-        endpoint_url = self.auth.get_endpoint(self.session,
-                                              service_type=self.service_type,
-                                              region_name=self.region_name,
-                                              interface=self.interface)
-        return {'auth_token': access_info.auth_token,
-                'auth_tenant_id': access_info.tenant_id,
-                'auth_user_id': access_info.user_id,
-                'endpoint_url': endpoint_url}
+        auth_info = {'auth_token': self.auth_token,
+                     'endpoint_url': self.endpoint_url}
+
+        # NOTE(jamielennox): This is the best we can do here. It will work
+        # with identity plugins which is the primary case but we should
+        # deprecate it's usage as much as possible.
+        try:
+            get_access = (self.auth or self.session.auth).get_access
+        except AttributeError:
+            pass
+        else:
+            auth_ref = get_access(self.session)
+
+            auth_info['auth_tenant_id'] = auth_ref.project_id
+            auth_info['auth_user_id'] = auth_ref.user_id
+
+        return auth_info
 
 
 # FIXME(bklei): Should refactor this to use kwargs and only
@@ -378,6 +365,7 @@ def construct_http_client(username=None,
                           tenant_id=None,
                           password=None,
                           auth_url=None,
+                          auth_plugin=None,
                           token=None,
                           region_name=None,
                           timeout=None,
@@ -389,15 +377,15 @@ def construct_http_client(username=None,
                           ca_cert=None,
                           service_type='network',
                           session=None,
-                          auth=None,
-                          roles=None):
+                          **kwargs):
 
     if session:
+        kwargs.setdefault('user_agent', 'python-neutronclient')
+        kwargs.setdefault('interface', endpoint_type)
         return SessionClient(session=session,
-                             auth=auth,
-                             interface=endpoint_type,
                              service_type=service_type,
-                             region_name=region_name)
+                             region_name=region_name,
+                             **kwargs)
     else:
         # FIXME(bklei): username and password are now optional. Need
         # to test that they were provided in this mode.  Should also
@@ -410,6 +398,7 @@ def construct_http_client(username=None,
                           auth_url=auth_url,
                           token=token,
                           endpoint_url=endpoint_url,
+                          auth_plugin=auth_plugin,
                           insecure=insecure,
                           timeout=timeout,
                           region_name=region_name,
@@ -417,5 +406,4 @@ def construct_http_client(username=None,
                           service_type=service_type,
                           ca_cert=ca_cert,
                           log_credentials=log_credentials,
-                          auth_strategy=auth_strategy,
-                          roles=roles)
+                          auth_strategy=auth_strategy)
